unit winmines;

{ By Paul Toth <tothpaul@multimania.com> }
{ http://www.multimania.com/tothpaul  }

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

Const
 maxx=30;
 maxy=24;

 delta:array[0..7,0..1] of integer=(
  (-1,-1),( 0,-1),(+1,-1),
  (-1, 0),        (+1, 0),
  (-1,+1),( 0,+1),(+1,+1)
 );

type
  TBoard=array[-1..maxx,-1..maxy] of byte;

  TWinMine = class(TCustomControl)
  private
   fBoard:TBoard;
   fFlags:TBoard;
   fTotalMines:integer;
   fMines:integer;
   fMaxX:integer;
   fMaxY:integer;
   fCells:integer;
   fEndOfGame:boolean;
   EOnNewGame:TNotifyEvent;
   EOnMineCount:TNotifyEvent;
   EOnWin:TNotifyEvent;
   EOnLose:TNotifyEvent;
   procedure SetCell(X,Y:integer; State:byte);
   procedure ClickCell(X,Y:integer);
  protected
   fSelected:boolean;
   fCellX,fCellY:integer;
   fCellState:byte;
   procedure Paint; override;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
   constructor Create(AOwner:TComponent); override;
   procedure NewGame(AWidth,AHeight,AMines:integer);
   procedure Restart;
   property Mines:integer read fMines;
   property EndOfGame:boolean read fEndOfGame;
  published
   property OnNewGame:TNotifyEvent read EOnNewGame write EOnNewGame;
   property OnMineCount:TNotifyEvent read EOnMineCount write EOnMineCount;
   property OnWin:TNotifyEvent read EOnWin write EOnWin;
   property OnLose:TNotifyEvent read EOnLose write EOnLose;
   property OnMouseDown;
   property OnMouseUp;
  end;

procedure Register;

implementation

{$R WINMINES.RES}
Var
 Images:HBitmap;

procedure Register;
begin
  RegisterComponents('MySoft', [TWinMine]);
end;

Constructor TWinMine.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  NewGame(16,16,40);
 end;

Procedure TWinMine.Paint;
 var
  dc:HDC;
  mx,my:integer;
 begin
  dc:=CreateCompatibleDC(Canvas.Handle);
  SelectObject(dc,Images);
  for mx:=0 to fmaxx-1 do
   for my:=0 to fmaxy-1 do
    BitBlt(Canvas.Handle,16*mx,16*my,16,16,DC,0,16*fFlags[mx,my],SRCCOPY);
  DeleteDC(dc);
 end;

Procedure TWinMine.NewGame(AWidth,AHeight,AMines:integer);
 begin
  fTotalMines:=AMines;
  fMaxX:=AWidth;
  fMaxY:=AHeight;
  Width:=16*AWidth;
  Height:=16*AHeight;
  Restart;
 end;

procedure TWinMine.Restart;
 var
  m,mx,my:integer;
 begin
  fEndOfGame:=False;
  fMines:=fTotalMines;
  fCells:=fMaxX*fMaxY;
  FillChar(fBoard,SizeOf(fBoard),0);
  FillChar(fFlags,SizeOf(fFlags),12);
  for m:=1 to fMines do begin
   Repeat
    mx:=Random(fmaxx);
    my:=Random(fmaxy);
   Until fBoard[mx,my]=0;
   fBoard[mx,my]:=9;
  end;
  for mx:=0 to fmaxx-1 do
   for my:=0 to fmaxy-1 do
    if fBoard[mx,my]<9 then begin
     m:=0;
     if fBoard[mx-1,my-1]=9 then inc(m);
     if fBoard[mx  ,my-1]=9 then inc(m);
     if fBoard[mx+1,my-1]=9 then inc(m);
     if fBoard[mx-1,my  ]=9 then inc(m);
     if fBoard[mx+1,my  ]=9 then inc(m);
     if fBoard[mx-1,my+1]=9 then inc(m);
     if fBoard[mx  ,my+1]=9 then inc(m);
     if fBoard[mx+1,my+1]=9 then inc(m);
     fBoard[mx,my]:=m;
    end;
  Invalidate;
  if Assigned(EOnNewGame) then EOnNewGame(Self);
  if Assigned(EOnMineCount) then EOnMineCount(Self);
 end;

procedure TWinMine.SetCell(X,Y:integer;State:byte);
 Var
  R:TRect;
 begin
  if (x<0)or(x>=fmaxx)or(y<0)or(y>=fmaxy) then exit;
  fFlags[X,Y]:=State;
  R.Left:=16*X; R.Right:=R.Left+16;
  R.Top :=16*Y; R.Bottom:=R.Top+16;
  InvalidateRect(Handle,@R,False);
 end;

procedure TWinMine.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 var
  m,f,mx,my:integer;
 begin
  inherited MouseDown(Button,Shift,X,Y);
  if fEndOfGame then exit;
  fCellX:=X div 16;
  fCellY:=Y div 16;
  fCellState:=fFlags[fCellX,fCellY];
  if Button=mbRight then begin // Flag
   if fCellState>=12 then begin // Unknow, Flag, ?
    inc(fCellState);
    if fCellState=15 then fCellState:=12;
    SetCell(fCellX,fCellY,fCellState);
    if (fCellState>12) then begin
     if fCellState=13 then dec(fMines) else inc(fMines);
     if Assigned(EOnMineCount) then EOnMineCount(Self);
    end;
   end;
  end else begin
   if ssDouble in Shift then begin
    if not (fCellState in [1..8]) then exit;
    f:=0;
    for m:=0 to 7 do begin
     case fFlags[fCellX+delta[m,0],fCellY+delta[m,1]] of
      13:inc(f); // flag
      14:exit;   // "?"
     end;
    end;
    if f=fCellState then for m:=0 to 7 do ClickCell(fCellX+delta[m,0],fCellY+delta[m,1])
   end else
   if fCellState=12 then begin
    fSelected:=True;
    SetCell(fCellX,fCellY,0);
    SetCaptureControl(Self);
   end;
  end;
 end;

procedure TWinMine.MouseMove(Shift: TShiftState; X, Y: Integer);
 begin
  if fSelected then begin
   SetCell(fCellX,fCellY,fCellState);
   fCellX:=X div 16;
   fCellY:=Y div 16;
   fCellState:=fFlags[fCellX,fCellY];
   if fCellState=12 then SetCell(fCellX,fCellY,0);
  end;
 end;

procedure TWinMine.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 begin
  inherited MouseUp(Button,Shift,X,Y);
  if fSelected then begin
   SetCaptureControl(nil);
   fSelected:=False;
   SetCell(fCellX,fCellY,fCellState);
   ClickCell(fCellX,fCellY);
  end;
 end;

Procedure TWinMine.ClickCell(X,Y:integer);
 var
  m:integer;
  mx,my:integer;
 begin
  if (X<0)or(X>=fmaxx)or(Y<0)or(Y>=fmaxy) then exit;
  if fFlags[X,Y]<>12 then exit;
  m:=fBoard[X,Y];
  if m=9 then begin
   SetCell(X,Y,10);
   for mx:=0 to fmaxx-1 do
    for my:=0 to fmaxy-1 do begin
     if (fBoard[mx,my]=9)and(fFlags[mx,my]=12) then SetCell(mx,my,9);
     if (fFlags[mx,my]>12)and(fBoard[mx,my]<>9) then SetCell(mx,my,11);
    end;
   fEndOfGame:=True; 
   if Assigned(EOnLose) then EOnLose(Self);
  end else begin
   dec(fCells);
   SetCell(X,Y,m);
   if m=0 then for m:=0 to 7 do ClickCell(x+delta[m,0],y+delta[m,1]);
   if (fCells=fTotalMines) then begin
    fEndOfGame:=True;
    if (Assigned(EOnWin)) then EOnWin(Self);
   end;
  end;
 end;


initialization
 Randomize;
 Images:=LoadBitmap(HInstance,'WINMINES');;
finalization
 DeleteObject(Images);

end.
